/*
   program : various

   $Source: /home/yee/C/RCS/array.c,v $
   $Author: yee $
   $Date: 1999/02/10 16:37:41 $
   $Revision: 9.21 $

   This file contains various useful utilities to manipulate arrays.

   Parameters :
     1) #define DEBUG if debug info is desired to be printed out
     2) #define MDVMALLOC if mdvmalloc() desired.
     3) #define MDMALLOC if mdmalloc() desired.
     4) #define MDFREE if mdfree() desired.
     5) #define MDIO if mdio() desired.
     6) #define MDACCESS if mdaccess() desired.
     7) for either mdio() and mdaccess()
        a) #define MD_ELEMENT (esize<=asize) (default: slower, more general)
        b) #define MD_ROW (esize==asize) 

   Bugs : none known

   To Do :
     1) rewrite mdfree() to handle ALLOC_CONTIGUOUS mode.
     2) ANSI'fy mdvmalloc().

   $Log: array.c,v $
   Revision 9.21  1999/02/10 16:37:41  yee
   *** empty log message ***

   Revision 9.20  1998/09/06 02:11:31  yee
   documentation fix.

   Revision 9.16  1997/04/30 17:44:43  yee
   fixed error in #define in mdio() & mdaccess().

   Revision 9.12  1997/04/24 17:03:05  yee
   added mdaccess()

   Revision 9.8  1997/02/19 03:29:42  yee
   bug fixes in mdfree().
   fixed bugs in mdio() that disallowed negative array indices.
   changed functions to use void pointers and not char pointers.

   Revision 9.1  1996/09/19 05:07:10  yee
   added CONTIGUOUS mode to code.  Major version number changed to maintain
   compatibility with MIPG versions.  Also deleted a number of function which
   provide a check to basic function (e.g. Fopen).

   Copyright (C) 1993,1994,1996,1997,1998 by Conway Yee.  All Rights Reserved.
 */

char array[]="$Id: array.c,v 9.21 1999/02/10 16:37:41 yee Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <varargs.h>

#include "array.h"
 
/*****************************************************************************/
/*****************************************************************************/

/* This function is a front end for mdmalloc() using varargs. */

#ifdef MDVMALLOC
void *mdvmalloc(ndim,size,mode,va_alist)

int ndim;             /* number of dimensions in array */
int size;             /* size (in bytes) of one element */
MODE mode;            /* operation options */
va_dcl

{
  va_list args;                 /* arguments consisting of array limits */
  int *lim;                     /* array w/ array limits for mdmalloc() */
  int i;                        /* counter */
  int begin;                    /* first element in dimension */
  int end;                      /* last element in dimension */
  void *ptr;                    /* pointer to malloc'ed array */

  va_start(args);
  lim=(int *)malloc(ndim*2*sizeof(int));

  if (lim == NULL) {
#ifdef DEBUG
    perror("mdvmalloc : unable to allocate limits array ");
#endif DEBUG
    return(NULL);
  } 
  else {
    for(i=0;i<ndim*2;) {        /* setup limits array */
      begin = va_arg(args,int);
      end = va_arg(args,int);
      if (end < begin) {
#ifdef DEBUG
        fprintf(stderr,"mdvmalloc : suspect array limits\n");
#endif DEBUG
        lim[i++] = end;
        lim[i++] = begin;
      }
      else {
        lim[i++]=begin;
        lim[i++]=end;
      }
    }
    va_end(ap);

    ptr=(void *)mdmalloc(ndim,size,mode,lim);  /* allocate array */
    free(lim);

    return (ptr);
  }
}
#endif MDVMALLOC

/*****************************************************************************/

/* This function is a Multi-Dimensional malloc.  This function can dynamically
   allocate an array of arbitrary size and dimensionality.  The starting index
   for each dimension is user defined.

   The argument lim is an array of integers which give the first and last
   index of each dimension.  The first pair of integers corresponds to the
   first dimension, the second pair, the second dimension, etc.

   This function is able to handle the trivial case of a zero dimension array.

   WARNING : all pointer types are assumed to be the same size. */

#if defined(MDMALLOC)
void *mdmalloc
(
 int ndim,      /* number of dimensions in array */
 int size,      /* size (in bytes) of one element */
 MODE mode,     /* operation options */
 int *lim       /* limits of each dimension */
)
{
  void **ptr;         /* pointer to malloc'ed array */
  int length;         /* number of elements in dimension */
  int i,j;            /* counters */

  static int m_alloc;  /* amount of alloc'ed memory left unused (in char) */
  static void *memory; /* alloc'ed memory-only used in ALLOC_CONTIGUOUS mode */

  if((mode==ALLOC_CONTIGUOUS)&&(m_alloc==0)) { /* allocate CONTIGUOUS memory */
    m_alloc = size;
    for(m_alloc=1,i=0;i<ndim;i++)
      m_alloc *= lim[i*2+1]-lim[i*2]+1;
    m_alloc *= size;
    memory = (void *)calloc(1,m_alloc);
  }

  if (ndim < 0) {                                  /* ndim must be positive! */
#ifdef DEBUG
    fprintf(stderr,"mdmalloc : illegal number of dimensions : %i\n",ndim);
#endif DEBUG
    return ((void *)MDMALLOC_ILLEGAL_DIMENSION);
  }

  if (ndim > 0)                 /* determine number of elements in dimension */
    length = lim[1]-lim[0]+1;
  else
    length = 1;
  if (length <= 0) {
#ifdef DEBUG
    fprintf(stderr,"mdmalloc : illegal dimension limits : [%i:%i]\n",lim[0],lim[1]);
#endif DEBUG
    return ((void *)MDMALLOC_ILLEGAL_DIMENSION);
  }

  if (ndim == 1) {                         /* allocate memory for array data */
    if (mode == ALLOC_CONTIGUOUS) {
      int mem=length*size;   /* memory used up */
      ptr = (void **)memory;
      memory += mem;
      m_alloc -= mem;
#ifdef DEBUG
      if(m_alloc<0)
        fprintf(stderr,"mdmalloc : ALLOC_CONTIGUOUS mode ran out of alloc'ed memory");
#endif DEBUG
    }
    else
      ptr = (void **)calloc(1,length*size);
    if (ptr == NULL) {
#ifdef DEBUG
        perror("mdmalloc : unable to malloc array : ");
#endif DEBUG
      return ((void *)ALLOC_ERROR);
    } 
    else {
      ptr = (void **)((unsigned int)ptr - lim[0]*size);
    }
  }
  else {                      /* allocate pointers to array */
    ptr = (void **)malloc(length*sizeof(void *));
    if (ptr == NULL) {
#ifdef DEBUG
      perror("mdmalloc : unable to malloc pointer to array ");
#endif DEBUG
      return ((void *)ALLOC_ERROR);
    } 
    else {
      ptr -= lim[0];                        /* set beginning limit */
      for(i=lim[0];i<=lim[1];i++) {         /* recurse to next dimension */
        ptr[i]=(void *)mdmalloc(ndim-1,size,mode,lim+2);
        if (ptr[i] <= (void *)NULL) {       /* error in malloc'ing */
          for(j=i;j>=lim[0];j--)            /* free prev. malloc'ed elements */
            mdfree(ndim-1,(void **)(lim+2),size,(int *)(ptr[j]));
          i=lim[1];                         /* exit loop */
        }
      }
    }
  }
  return ((void *)ptr);
}
#endif defined(MDMALLOC)

/*****************************************************************************/

/* This function frees the memory allocated when mdmalloc() is used.

   WARNING: this function has not been tested. */

#ifdef MDFREE
void mdfree
(
int ndim,   /* number of dimensions in array */
void *ptr,  /* pointer to array to be freed */
int size,   /* size (in bytes) of one element */
int *lim    /* limits of each dimension */
)
{
  int i;  /* counter */

  if(ndim < 0) {
#ifdef DEBUG
    fprintf(stderr,"mdfree : WARNING unable to free array : illegal number of dimensions : %i : ",ndim);
#endif DEBUG
    perror(NULL);
  } else
    if(ndim == 1)                                         /* free array data */
      free((void *)((unsigned int)ptr+lim[0]*size));
    else {                                            /* free array pointers */
      for(i=lim[0];i<=lim[1];i++)               /* recurse to next dimension */
        mdfree(ndim-1,((void **)ptr)[i],size,lim+2);

      free((void *)ptr+lim[0]);            /* free pointer to this dimension */
    }
}
#endif MDFREE

/*****************************************************************************/

/* MultiDimensional Input Output

   This function handles multi-dimensional input/output of data to/from a
   file.  In order to handle the case of putting real floats into a complex
   array, this function is written to allow the array elements to be of
   a different size than the elements read.

   WARNING : function assumes that all arrays are pointer to pointers! */

#ifdef MDIO
PUBLIC void *mdio
(
 FILE *fptr,          /* file used for input/output */
 unsigned int ndim,   /* number of dimensions in array */
 unsigned int esize,  /* size (in bytes) of one element read from fptr */
 unsigned int asize,  /* size (in bytes) of one element in array a */
 void *a,             /* array used for input/output */
 size_t (*IO)(),      /* fread() if input, fwrite() if output */
 int *lim             /* limits in each dimension */
)
{
  int i; /* counter */

  if(ndim==1) {
    a=(void **)((unsigned int)a+lim[0]*asize);
#ifdef MD_ROW            
                         /* read one row at a time. use iff (esize == asize) */
    (*IO)((void *)a,esize,lim[1]-lim[0]+1,fptr);
#elif defined(MD_ELEMENT)
                         /* read one element at a time. for (esize <= asize) */
    for(i=lim[0];i<=lim[1];i++,a=(void **)((unsigned int)a+asize))
      (*IO)((void *)a,esize,1,fptr);
#else
#error MD_ROW_NOR_MD_ELEMENT_DEFINED
#endif MD_ROW&&MD_ELEMENT
  } else
    for(i=lim[0];i<=lim[1];i++)
      mdio(fptr,ndim-1,esize,asize,((void **)a)[i],IO,lim+2);

  return((void *)a);
}
#endif MDIO

/*****************************************************************************/

/* MultiDimensional array element access.

   This function takes each element of a multi-dimensional array and hands it
   to a function func() which does whatever it wants to element.

   WARNING : function assumes that all arrays are pointer to pointers! */

#ifdef MDACCESS
PUBLIC void mdaccess
(
 unsigned int ndim,   /* number of dimensions in array */
 unsigned int size,   /* size (in bytes) of one element in array a */
 void *a,             /* array used for input/output */
 void (*func)(),      /* function acting on array element */
 int *lim,            /* limits in each dimension */
 int *IntFree,        /* integer degrees of freedom */
 double *FloatFree    /* floating point degrees of freedom */
)
{
  int i; /* counter */

#ifdef MD_ELEMENT
  if(ndim==0)
    (*func)(a,IntFree,FloatFree);
  else
#endif MD_ELEMENT
    if(ndim==1) {
      a=(void **)((unsigned int)a+lim[0]*size);
      for(i=lim[0];i<=lim[1];i++,a=(void **)((unsigned int)a+size)) {
#ifdef MD_ROW
        (*func)(a);
                         /* read one element at a time. for (esize <= asize) */
#elif defined(MD_ELEMENT)
        mdaccess(ndim-1,size,(void **)a,func,lim+2,IntFree,FloatFree);
#else
#error MD_ROW_NOR_MD_ELEMENT_DEFINED
#endif MD_ROW&&MD_ELEMENT
      }
    } else
      for(i=lim[0];i<=lim[1];i++)
        mdaccess(ndim-1,size,((void **)a)[i],func,lim+2,IntFree,FloatFree);
}
#endif MDACCESS
